from jinja2 import Environment, PackageLoader, select_autoescape

# Filters


def rgb(rgb_color):
    #
    # NOTE: Inkscape does not support rgb color!!!
    #
    # Convert RGB to hex

    hex_color = "#"
    for val in rgb_color[:3]:
        hex_color += ("0" + hex(val).split("x")[-1])[-2:]
    return hex_color


env = Environment(
    loader=PackageLoader("pinout", "templates"),
    autoescape=select_autoescape(["html", "xml"]),
    trim_blocks=True,
    lstrip_blocks=True,
)

# Add filters to Jinja env
env.filters["rgb"] = rgb

# Base SVG elements
svg = env.get_template("svg.svg")
svg_group = env.get_template("group.svg")
svg_image = env.get_template("image.svg")
svg_rect = env.get_template("rect.svg")

# Component SVG templates (made from multiple svg elements)
svg_label = env.get_template("label.svg")
svg_annotation = env.get_template("annotation.svg")

svg_path = env.get_template("path.svg")
svg_text = env.get_template("text.svg")
svg_textblock = env.get_template("textblock.svg")



import copy
from . import file_manager
from .templates import svg_group
from .elements import Element, SVG, Coords, BoundingBox, BoundingCoords
from . import elements as elem


class Component(SVG):
    """Container object that manages child Components and/or Elements as a group.

    :param children: Components and/or elements, defaults to None
    :type children: SVG subtypes, optional
    """

    config = {}

    def __init__(self, children=None, *args, **kwargs):
        """[summary]"""
        self.children = []
        super().__init__(*args, **kwargs)

    @property
    def bounding_coords(self):
        """Coordinates of the components's bounding rectangle.

        :return: (x_min, y_min, x_max, y_max)
        :rtype: BoundingCoords (namedtuple)
        """
        # Untransformed bounding coords
        x_min = y_min = x_max = y_max = 0
        for child in self.children:
            try:
                child_coords = child.bounding_coords
                x_min = min(x_min, child_coords.x_min)
                y_min = min(y_min, child_coords.y_min)
                x_max = max(x_max, child_coords.x_max)
                y_max = max(y_max, child_coords.y_max)
            except AttributeError:
                # The child has no bounding_coords.
                pass

        x_min, x_max = sorted(
            [(self.x + x_min) * self.scale.x, (self.x + x_max) * self.scale.x]
        )
        y_min, y_max = sorted(
            [(self.y + y_min) * self.scale.y, (self.y + y_max) * self.scale.y]
        )

        return BoundingCoords(x_min, y_min, x_max, y_max)

    @property
    def bounding_rect(self):
        """Components's coordinates and size.

        :return: (x, y, width, height)
        :rtype: BoundingBox (namedtuple)
        """
        x_min, y_min, x_max, y_max = self.bounding_coords
        return BoundingBox(x_min, y_min, x_max - x_min, y_max - y_min)

    @property
    def width(self):
        """Calculated width that encompasses all child elements

        :return: value representing width in pixels
        :rtype: int
        """
        try:
            x_min, y_min, x_max, y_max = self.bounding_coords
            return x_max - x_min
        except ValueError:
            # Component has no children with bounding_coords
            return 0

    @property
    def height(self):
        """Calculated height that encompasses all child elements

        :return: value representing height in pixels
        :rtype: int
        """
        try:
            x_min, y_min, x_max, y_max = self.bounding_coords
            return y_max - y_min
        except ValueError:
            # Component has no children with bounding_coords
            return 0

    @property
    def scale(self):
        """See :func:`~pinout.elements.SVG.scale`"""
        return self._scale

    @scale.setter
    def scale(self, value):
        self._scale = value
        for child in self.children:
            if issubclass(type(child), Element):
                child.scale = self.scale

    def add(self, instance):
        """Add an instance to the component's children.

        :param instance: Component or Element
        :type instance: Component or Element
        :return: instance attribute
        :rtype: Component or Element
        """
        self.children.append(instance)
        return instance

    @staticmethod
    def patch_config(source, patch):
        """Recursively update source with patch dict items."""
        try:
            for key, val in patch.items():
                if type(val) == dict:
                    source.setdefault(key, {})
                    Component.patch_config(source[key], patch[key])
                else:
                    source[key] = val
        except KeyError:
            # patch has no items
            pass
        return source

    def render(self):
        """Render Component, and children, as SVG markup.

        :return: SVG markup of component including all children.
        :rtype: str
        """
        output = ""
        for child in self.children:
            output += child.render()
        return svg_group.render(
            x=self.x,
            y=self.y,
            tag=self.tag,
            content=output,
            scale=self.scale,
        )


class PinLabelSet(Component):
    """Add rows of PinLabels to a 'header' of pins

    :param offset: Offset of the first PinLabelRow from the origin
    :type offset: (x, y) tuple
    :param labels: List of PinLabelRows
    :type labels: List
    :param pitch: Offset between each PinLabelRow, defaults to (1, 1)
    :type pitch: (x, y) tuple, optional
    """

    def __init__(self, offset, labels, pitch=(1, 1), *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Extract scale and update x and y of pinlabelset instance
        offset, self.scale = self.extract_scale(offset)
        self.x = self.x * self.scale.x
        self.y = self.y * self.scale.y

        pitch = Coords(*pitch)

        # Create a Component for each row in 'labels'
        for i, label_list in enumerate(labels):
            pin_x = pitch.x * i * self.scale.x
            pin_y = pitch.y * i * self.scale.y
            row_offset = offset
            if pitch[1] == 0:
                # Horizontal pinset require label_rows to offset vertically
                row_offset = Coords(offset.x - pin_x, offset.y + abs(pin_x))

            row = self.add(
                Component(
                    x=pin_x + row_offset.x,
                    y=pin_y + row_offset.y,
                )
            )

            # Create a leaderline
            leaderline_config = copy.deepcopy(self.config["leaderline"])
            vertical_move = f"V {row_offset.y}" if row_offset.y != 0 else ""
            horizontal_move = (
                f"H {row_offset.x + pitch.x * i * self.scale.x}"
                if row_offset.x != 0
                else ""
            )

            definition = f"M {pin_x} {pin_y} {vertical_move} {horizontal_move}"

            leaderline = self.add(
                elem.Path(
                    definition=definition,
                    config=leaderline_config,
                )
            )

            # Add labels to row
            for j, label in enumerate(label_list):
                label = dict(zip(("text_content", "tag", "config"), label))

                # Copy config and patch with supplied config
                label_config = copy.deepcopy(self.config)
                self.patch_config(label_config, label.get("config", {}))
                # Patch config with tag styles
                tag_color = Component.config["tags"][label["tag"]]["color"]
                self.patch_config(
                    label_config,
                    {
                        "label": {"rect": {"fill": tag_color}},
                        "leaderline": {"stroke": tag_color},
                    },
                )

                # Match leaderline to first label tag color
                if j == 0:
                    leaderline.config["stroke"] = tag_color

                # add label's leaderline
                label_offset = Coords(*label_config["offset"])
                self.patch_config(
                    label_config,
                    {"leaderline": {"stroke": tag_color}},
                )
                definition = f"M {row.width} 0 h {label_offset.x}"
                row.add(
                    elem.Path(
                        x=row.width,
                        y=0,
                        width=label_offset.x,
                        height=self.config["leaderline"]["stroke_width"],
                        scale=self.scale,
                        definition=definition,
                        config=label_config["leaderline"],
                    )
                )

                row.add(
                    elem.Label(
                        text_content=label["text_content"],
                        x=row.width,
                        y=-label_config["label"]["rect"]["height"] / 2,
                        width=label_config["label"]["rect"]["width"],
                        height=label_config["label"]["rect"]["height"],
                        scale=self.scale,
                        config=label_config["label"],
                    )
                )


class Legend(Component):
    """Provide a colour coded legend to describe pin labels.

    :param categories: List of tags to include in legend
    :type categories: [<tag1>, <tag2>, ...] List
    """

    def __init__(self, categories, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # NOTE: pad.y sets the text **baseline**
        pad = Coords(*self.config["padding"])
        row_height = self.config["row_height"]
        swatch_size = row_height * 2 / 3
        categories = categories or Component.config["tags"].keys()
        for i, tag in enumerate(categories):
            entry = self.add(
                Component(x=pad.x, y=pad.y + row_height * i, tag=self.config["tag"])
            )
            entry.add(
                elem.Text(
                    Component.config["tags"][tag]["title"],
                    x=swatch_size * 2,
                    y=0,
                    width=self.config["rect"]["width"],
                    height=row_height,
                    config=self.config["text"],
                )
            )

            # Create icon based on pinlabel config
            pinlabel_config = copy.deepcopy(Component.config["pinlabel"])
            tag_color = Component.config["tags"][tag]["color"]
            pinlabel_patch = {
                "offset": (-swatch_size / 2, 0),
                "label": {
                    "rect": {
                        "fill": tag_color,
                        "height": swatch_size,
                        "width": swatch_size,
                        "rx": pinlabel_config["label"]["rect"]["rx"],
                    },
                },
                "leaderline": {
                    "stroke": tag_color,
                },
            }
            self.patch_config(pinlabel_config, pinlabel_patch)

            entry.add(
                elem.Rect(
                    x=0,
                    y=-pinlabel_config["label"]["rect"]["height"] / 2
                    - self.config["text"]["size"] / 2,
                    width=swatch_size,
                    height=swatch_size,
                    config=pinlabel_config["label"]["rect"],
                )
            )

            definition = f"M {swatch_size} {-pinlabel_config['label']['text']['size'] / 2} h {swatch_size/2}"
            entry.add(
                elem.Path(
                    definition=definition,
                    x=swatch_size,
                    y=-pinlabel_config["label"]["text"]["size"] / 2,
                    width=swatch_size / 2,
                    height=pinlabel_config["leaderline"]["stroke_width"],
                    config=pinlabel_config["leaderline"],
                )
            )

        # Add an panel *behind* component
        self.config["rect"]["height"] = (
            row_height * len(categories) + pad.y - self.config["text"]["size"]
        )
        self.children.insert(
            0,
            elem.Rect(
                x=0,
                y=0,
                width=self.config["rect"]["width"],
                height=self.config["rect"]["height"],
                scale=self.scale,
                config=self.config["rect"],
            ),
        )


class Annotation(Component):
    """Add text with a leaderline styled as an annotation.

    :param text_content: Annotation text
    :type text_content: String or List
    """

    def __init__(self, text_content, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Extract scale from offset and update x and y
        offset, self.scale = self.extract_scale(self.config["offset"])

        # override scale if explicitly provided
        if "scale" in kwargs.keys():
            self.scale = Coords(*kwargs["scale"])

        self.x = self.x * self.scale.x
        self.y = self.y * self.scale.y

        label_padding = Coords(*self.config["label"]["padding"])

        # Attempt to split on '\n' and convert to list
        if type(text_content) == str:
            text_content = text_content.split("\n")

        # Calculate label dimensions
        line_height = self.config["label"]["text"]["line_height"]
        font_height = self.config["label"]["text"]["size"]
        top_padding = label_padding.y - (line_height - font_height)
        label_height = (
            len(text_content) * self.config["label"]["text"]["line_height"]
            + label_padding.y
            + top_padding
        )
        self.config["label"]["rect"]["height"] = label_height

        # label required nudging to align with leaderline
        stroke_shim = self.config["leaderline"]["stroke_width"] / 2

        # Annotation label
        # shift label if flipped
        label_translate_y = label_height if self.scale.y == 1 else 0
        label = self.add(
            Component(
                tag="anno_label",
                x=offset.x - stroke_shim,
                y=offset.y - label_translate_y + stroke_shim,
            )
        )
        # Add background rect to label
        label.add(
            elem.Rect(
                y=0,
                width=self.config["label"]["rect"]["width"],
                height=self.config["label"]["rect"]["height"],
                config=self.config["label"]["rect"],
            )
        )
        # Add textblock to label
        label.add(
            elem.TextBlock(
                text_content,
                x=label_padding.x,
                y=top_padding,
                width=self.config["label"]["rect"]["width"] - label_padding.x * 2,
                height=self.config["label"]["rect"]["height"] - label_padding.y,
                config=self.config["label"],
                scale=self.scale,
            )
        )

        # Leaderline
        # leaderline rect
        leaderline_rect = self.add(
            elem.Rect(
                x=-self.config["leaderline"]["rect"]["width"] / 2,
                y=-self.config["leaderline"]["rect"]["height"] / 2,
                width=self.config["leaderline"]["rect"]["width"],
                height=self.config["leaderline"]["rect"]["height"],
                config=self.config["leaderline"]["rect"],
            )
        )

        # leaderline start location at edge of leaderline_rect
        start_y = 0
        start_x = 0
        if offset.y > leaderline_rect.height / 2:
            start_y = leaderline_rect.height / 2
        elif -leaderline_rect.height / 2 < offset.y < leaderline_rect.height / 2:
            start_x = leaderline_rect.width / 2

        # leaderline path
        vertical_move = f"V {offset.y}" if offset.y != 0 else ""
        horizontal_move = f"H {offset.x + label.width - stroke_shim}"
        label_width = self.config["label"]["rect"]["width"]
        horizontal_move = f"H {offset.x + label.width - stroke_shim}"

        path_definition = f"M {start_x} {start_y} {vertical_move} {horizontal_move}"

        self.add(elem.Path(path_definition, config=self.config["leaderline"]))



###########################################
#
# Example script to build a pinout diagram
# with some annotation labels
#
###########################################

from pinout.diagram import Diagram
from pinout.components import Component


# Create a new diagram
diagram = Diagram()


# annotations_01
# --------------

# Create a child diagram
annotations_01 = diagram.add(Diagram())

# Supply config settings.
annotations_01.add_config("annotations_config.yaml")

# Add the hardware board image
annotations_01.add_image("hardware_board.svg", width=220, height=300)

# Add some annotations.
annotations_01.add_annotation(["Micro", "USB-C"], x=110, y=24)
annotations_01.add_annotation("Onboard LED", x=87, y=85)
annotations_01.add_annotation("STM32L0 \nSingle core Cortex-M0+", x=110, y=185)
annotations_01.add_annotation("Additional 4 pin header", x=110, y=284)

# If your needs are simple and some care is taken with the config this might be adequate, but probably not ideal for most situations.


# annotations_02
# --------------

# This diagram will be position below 'annotations_01'
annotations_02 = diagram.add(Diagram(y=380))
annotations_02.add_config("annotations_config.yaml")
annotations_02.add_image("hardware_board.svg", width=220, height=300)

# Config can be supplied per annotation in the form of a dict matching the config.yaml annotation entry
annotations_02.add_annotation(
    "Onboard LED",
    x=87,
    y=85,
    config={
        "offset": (163, 40),
        "leaderline": {
            "rect": {
                "width": 36,
                "height": 36,
                "fill_opacity": 0,
            }
        },
    },
)
annotations_02.add_annotation(
    "STM32L0 \nSingle core Cortex-M0+", x=110, y=185, config={"offset": (-140, 0)}
)
annotations_02.add_annotation(
    "Additional 4 pin header",
    x=110,
    y=284,
    config={
        "leaderline": {"rect": {"width": 134, "height": 46, "fill_opacity": 0}},
        "offset": (140, -46),
    },
)

# As an alternative component config can be documented in the YAML config file and reference here.
annotations_02.add_annotation(
    ["Micro", "USB-C"],
    x=110,
    y=24,
    scale=(-1, -1),
    config=annotations_02.config["anno_2_usb"],
)
# NOTE: The sign of offset can be overridden by supplying a 'scale=(<+/-1>,<+/-1>)'


# annotations_03
# --------------

# This diagram will be position below 'annotations_02'
annotations_03 = diagram.add(Diagram(y=710))
annotations_03.add_config("annotations_config.yaml")
annotations_03.add_image("hardware_board.svg", width=220, height=300)

# Patch config["annotation"] with config["anno_3_defaults"].
annotations_03.patch_config(
    annotations_03.config, {"annotation": annotations_03.config["anno_3_defaults"]}
)

# USB connection
annotations_03.add_annotation(
    "Micro USB-C", x=110, y=24, config=annotations_03.config["anno_3_usb"]
)

# Microprocessor
orange = [246, 148, 30]
annotations_03.add_annotation(
    "STM32L0 \n Single core Cortex-M0+ \n32kb Flash \n USART, SPI, I2C, USB",
    x=110,
    y=171,
    config=annotations_03.config["anno_3_mc"],
)

# Onboard LED
annotations_03.add_annotation(
    "Onboard LED", x=87, y=85, config=annotations_03.config["anno_3_led"]
)

# Pin header
annotations_03.add_annotation(
    "Additional 4 pin header",
    x=110,
    y=284,
    config=annotations_03.config["anno_3_header"],
)


diagram.export("annotated_diagrams.svg", overwrite=True)


###########################################
#
# Example script to build a pinout diagram
# Includes examples of all features
#
###########################################

from pinout.diagram import Diagram

# Custom pinlabel config
long_label_cfg = {"offset": (66, 0), "label": {"rect": {"width": 108}}}

# Pin label information grouped into dicts that render sets of labels
pindata = [
    # Left side vetical header
    {
        "x": 16,
        "y": 105,
        "pitch": (0, 30),
        "offset": (-76, 0),
        "labels": [
            [("0", "gpio"), ("Lb0a", "gpio"), ("Lb03", "comms")],
            [("1", "gpio"), ("Lc1a", "comms")],
            [("2", "gpio"), ("Lc2a", "analog"), ("Lc2b", "led")],
            [("RESET", "pwr", long_label_cfg)],
        ],
    },
    # Right side vetical header
    {
        "x": 204,
        "y": 105,
        "pitch": (0, 30),
        "offset": (76, 0),
        "labels": [
            [("Vcc", "pwr", long_label_cfg)],
            [("GND", "pwr", long_label_cfg)],
            [("3", "gpio"), ("Lc2a", "analog"), ("Lc2b", "led")],
            [("4", "gpio"), ("Lc3a", "analog")],
        ],
    },
    # Bottom horizontal header (Left half)
    {
        "x": 65,
        "y": 284,
        "pitch": (30, 0),
        "offset": (-125, 40),
        "labels": [
            [("3", "gpio"), ("Lc2a", "analog"), ("Lc2b", "led")],
            [("4", "gpio"), ("Lc2a", "pwm"), ("Lc2b", "led")],
        ],
    },
    # Bottom horizontal header (Right half)
    {
        "x": 155,
        "y": 284,
        "pitch": (-30, 0),
        "offset": (125, 40),
        "labels": [
            [("6", "gpio"), ("Lc3a", "analog")],
            [("5", "gpio"), ("Lc2a", "pwm"), ("Lc2b", "led")],
        ],
    },
    # Left interior pin
    {
        "x": 47,
        "y": 80,
        "offset": (-107, -100),
        "labels": [
            [("a", "led"), ("AUX_a", "analog")],
        ],
    },
    # Right interior pin
    {
        "x": 62,
        "y": 95,
        "pitch": (30, 0),
        "offset": (-122, -145),
        "labels": [
            [("b", "led"), ("AUX_b", "pwm"), ("PWM", "pwm")],
        ],
    },
]


# Create a new diagram object
diagram = Diagram(tag="full_sample")

# Override default config with user supplied values
# This file includes 'tag' information to style pinlabels
diagram.add_config("full_sample_config.yaml")

# Add and embed external image
diagram.add_image("full_sample_hardware.png", width=220, height=300, embed=True)

# Add pinlabel legend

diagram.add_legend(
    x=-234, y=380, categories=["analog", "comms", "gpio", "led", "pwm", "pwr"]
)

# Add pinlabels
for data in pindata:
    diagram.add_pinlabelset(**data)

# Add an annotation label
diagram.add_annotation(
    "USB-C connector  \nHost/Device functionality",
    x=110,
    y=20,
    config={"offset": (114, -69)},
)

# Export final SVG diagram
diagram.export("full_sample_diagram.svg", True)


import copy
import os
from collections import namedtuple
from pathlib import Path
from . import file_manager
from .elements import Image, Coords, BoundingBox
from .components import Component, Legend, PinLabelSet, Annotation
from .templates import svg


class Diagram(Component):
    """The base class that makes up a *pinout* diagram."""

    def __init__(self, config=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Load default config and patch with any supplied patch
        Component.config = file_manager.load_config()

        # Patch config with user suppled config.
        self.config = Component.config
        try:
            self.patch_config(self.config, config)
        except AttributeError:
            # No config supplied by user
            pass

    def add_config(self, path=None):
        """Add a configuration file to the diagram."""
        self.patch_config(self.config, file_manager.load_config(path))

    def add_image(self, path, *args, embed=False, **kwargs):
        """Associate a PNG, JPG or SVG formatted image to the diagram."""
        self.add(Image(path, *args, embed=embed, **kwargs))

    def add_legend(self, *args, categories=None, **kwargs):
        """Add a pinlabel legend to the diagram."""
        config = copy.deepcopy(self.config["legend"])
        kwargs["config"] = self.patch_config(config, kwargs.get("legend", {}))
        self.add(Legend(categories, *args, **kwargs))

    def add_pinlabelset(self, *args, **kwargs):
        """Add a pinlabels to a 'header' of pins in the diagram."""
        config = copy.deepcopy(self.config["pinlabel"])
        kwargs["config"] = self.patch_config(config, kwargs.get("config", {}))
        self.add(PinLabelSet(*args, **kwargs))

    def add_annotation(self, text_content, *args, **kwargs):
        """Add an annotation to the diagram."""
        config = copy.deepcopy(self.config["annotation"])
        kwargs["config"] = self.patch_config(config, kwargs.get("config", {}))
        self.add(Annotation(text_content, *args, **kwargs))

    def export(self, path, overwrite=False):
        """Output the diagram in SVG format.

        :param path: File location and name
        :type path: string
        :param overwrite: Overwrite existing file of same path, defaults to False
        :type overwrite: bool, optional
        """
        # Create export location and unique filename if required
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        if not overwrite:
            path = file_manager.unique_filepath(path)
        path.touch(exist_ok=True)

        output = ""

        # Render all components
        for c in self.children:
            output += c.render()

        # Add padding to viewbox
        padding = Coords(*self.config["diagram"]["padding"])
        x, y, w, h = self.bounding_rect
        viewbox = BoundingBox(
            x - padding.x, y - padding.y, w + padding.x * 2, h + padding.y * 2
        )

        # Render final SVG file
        path.write_text(
            svg.render(
                x=0,
                y=0,
                width=self.width,
                height=self.height,
                viewbox=viewbox,
                tag=self.tag,
                content=output,
            )
        )
        print(f"'{path}' exported successfully.")



# File Manager: common file manipulation functions
from pathlib import Path
import pkg_resources
import yaml


def unique_filepath(filepath):
    """Generate a unique filename to ensure existing files are not overwritten.

    :param filepath: File-name, including path, to the file location
    :type filepath: str
    :return: Unique filepath (filename and path to file location)
    :rtype: str
    """
    filepath = Path(filepath)
    suffix = filepath.suffix
    name = filepath.name[: -len(filepath.suffix)]
    path = filepath.parent
    count = 0

    while filepath.is_file():
        count += 1
        filepath = Path("{}/{}_{}{}".format(path, name, count, suffix))
    return filepath


def export_file(content, path, overwrite=False):
    """Create a file and export content to it. If overwrite is False a new filename is generated.

    :param content: Text data to be written to file
    :type content: str
    :param path: Path, including filename, for the file to be created
    :type path: str
    :param overwrite: Dictate if an existing file should be overwritten. If False 'path' is amended to be a unique filename, defaults to False
    :type overwrite: bool, optional
    :return: Actual path the file was written to. Note: this may differ from supplied 'path' if 'overwrite' is False)
    :rtype: str
    """
    if isinstance(path, str):
        path = Path(path)
    if not overwrite:
        path = unique_filepath(path)
    path.touch(exist_ok=True)
    path.write_text(content)
    return path


def load_config(path=None):
    """Load a config file either from a user nominated location or, if no path is supplied, load the default config file from the *pinout* package.

    :param path: [description], defaults to None
    :type path: [type], optional
    :return: [description]
    :rtype: [type]
    """
    if path is None:
        # Load default settings
        path = "resources/config.yaml"
        return yaml.safe_load(
            pkg_resources.resource_string(__name__, path).decode("utf-8")
        )
    # Load user supplied settings
    path = Path(path)
    with path.open() as f:
        return yaml.safe_load(f)


def duplicate(resource_name):
    """Duplicate resources from *pinout* package.

    A variety of resources can be copied via this function. Access is also available via the command line::

        py -m pinout.file_manager --duplicate <resource_name>

    Where :code:`<resource_name>` can be any one of the following resources:

        + config
        + quick_start
        + full_sample
        + annotations

    :param resource_name: Name of resource to duplicate.
    :type resource_name: string
    """

    resources = {
        "config": [(".", "config.yaml")],
        "quick_start": [
            ("quick_start", "quick_start_config.yaml"),
            ("quick_start", "quick_start_hardware.png"),
            ("quick_start", "quick_start_pinout.py"),
        ],
        "full_sample": [
            ("full_sample", "full_sample_config.yaml"),
            ("full_sample", "full_sample_hardware.png"),
            ("full_sample", "full_sample_pinout.py"),
        ],
        "annotations": [
            ("annotations", "annotations.py"),
            ("annotations", "annotations_config.yaml"),
            ("shared", "hardware_board.svg"),
        ],
    }

    resource_package = "pinout"
    for path in resources[resource_name]:
        resource_path = "/".join(("resources", *path))
        data = pkg_resources.resource_string(resource_package, resource_path)
        filename = path[-1]
        with open(filename, "wb") as f:
            f.write(data)
        print(f"{filename} duplicated.")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--duplicate",
        action="store",
        choices=["quick_start", "full_sample", "config", "annotations"],
        help="duplicate pinout resources",
    )
    args = parser.parse_args()
    if args.duplicate:
        duplicate(args.duplicate)


""" Elements: Basic building blocks for pinout diagrams
"""

# Elements
# Base building blocks to create SVG components
import base64
import pathlib
import uuid
from collections import namedtuple
from .templates import (
    svg_group,
    svg_image,
    svg_label,
    svg_path,
    svg_rect,
    svg_text,
    svg_textblock,
)


BoundingBox = namedtuple("BoundingBox", ("x y w h"))
BoundingCoords = namedtuple("BoundingCoords", ("x_min y_min x_max y_max"))
Coords = namedtuple("Coords", ("x y"))


class ClassMethodMissing(Exception):
    """ An element is missing an expected method """

    pass


class SVG:
    """Common base for all SVG entities that ultimately have a graphical representation.

    :param x: Coordinate position in the 'x' direction, defaults to 0
    :type x: int, optional
    :param y: Coordinate position in the 'y' direction, defaults to 0
    :type y: int, optional
    :param scale: Primarily used to define direction of width, height, and coordinates, from an origin. Defaults to (1, 1)
    :type scale: tuple, optional
    :param rotation:  (***Currently NOT implemented***) Rotation around an origin, defaults to 0
    :type rotation: int, optional
    :param tag: Associate an entity for application of predefined attributes/styles, defaults to ""
    :type tag: str, optional
    :param config: Directly supply attributes to an entity. Alternative to assigning via 'tag' or in the default config. Defaults to None
    :type config: dict, optional
    """

    def __init__(self, x=0, y=0, scale=(1, 1), rotation=0, tag="", config=None):
        """Create a new SVG object."""
        self.config = config
        self.rotation = rotation
        self._scale = Coords(*scale)
        self.tag = tag
        self.x = x
        self.y = y

    @property
    def scale(self):
        """Tuple representing orientation of entity."""
        # NOTE: set here as a property for override by Component
        return self._scale

    @scale.setter
    def scale(self, value):
        """Scale setter property - overridden by Component

        :param value: (x, y) where x and y are either 1 or -1
        :type value: tuple
        """
        # NOTE: set here for override by Component
        self._scale = value

    @staticmethod
    def extract_scale(coords):
        """Separate and scale information from *coords*.

        :param coords: (x, y) coordinates or (width, height) dimensions.
        :type coords: tuple
        :return: 'coords' parameter with absolute values and scale Coords.
        :rtype: (Coords, Coords), namedtuples
        """

        scale = Coords(*[i / abs(i) if i != 0 else 1 for i in coords])
        abs_coords = Coords(*[abs(i) for i in coords])
        return (abs_coords, scale)


class Element(SVG):
    """Fundamental building blocks that render SVG markup.

    :param width: Width of the rendered element, defaults to 0
    :type width: int, optional
    :param height: Height of the rendered element, defaults to 0
    :type height: int, optional

    """

    def __init__(self, width=0, height=0, *args, **kwargs):
        """Create a new Element"""
        self.width = width
        self.height = height
        super().__init__(*args, **kwargs)

    @property
    def bounding_coords(self):
        """Coordinates of the element's bounding rectangle.

        :return: (x_min, y_min, x_max, y_max)
        :rtype: BoundingCoords (namedtuple)
        """
        #########################################################
        #
        #        !!! IMPORTANT !!!
        #
        # ----- * Scale: Expected default behaviour * -----
        # Effects of 'scale' on Elements may vary.
        # Default behaviour expects an elements bounding coords
        # is unchanged when scale is applied to an element.
        #
        # *IF* an element's bounding coords changes this method must
        # be written specifically.
        #
        #########################################################

        x_min, x_max = sorted(
            [self.x * self.scale.x, (self.x + self.width) * self.scale.x]
        )
        y_min, y_max = sorted(
            [self.y * self.scale.y, (self.y + self.height) * self.scale.y]
        )

        x_min, x_max = sorted([self.x, self.x + self.width])
        y_min, y_max = sorted([self.y, self.y + self.height])

        return BoundingCoords(x_min, y_min, x_max, y_max)

    @property
    def bounding_rect(self):
        """Element's coordinates and size.

        :return: (x, y, width, height)
        :rtype: BoundingBox (namedtuple)
        """
        x_min, y_min, x_max, y_max = self.bounding_coords
        return BoundingBox(x_min, y_min, x_max - x_min, y_max - y_min)

    def render(self):
        """Create SVG markup of the element"""
        raise ClassMethodMissing(f"{self} requires a 'render' method.")


class Image(Element):
    """Associate a PNG, JPG or SVG formatted image to the diagram.

    :param href: path to image
    :type href: string
    :param embed: Embed image in rendered SVG file, defaults to False
    :type embed: bool, optional
    """

    def __init__(self, href, embed=False, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.href = href
        self.embed = embed

    @property
    def bounding_coords(self):
        """Coordinates of element boundaries"""

        x_min, x_max = sorted(
            [self.x * self.scale.x, (self.x + self.width) * self.scale.x]
        )
        y_min, y_max = sorted(
            [self.y * self.scale.y, (self.y + self.height) * self.scale.y]
        )

        # x_min, x_max = sorted([self.x, self.x + self.width])
        # y_min, y_max = sorted([self.y, self.y + self.height])

        return BoundingCoords(x_min, y_min, x_max, y_max)

    def render(self):
        """Render SVG markup either linking or embedding an image.

        :return: SVG markup
        :rtype: string
        """
        media_type = pathlib.Path(self.href).suffix[1:]
        path = pathlib.Path(self.href)
        if self.embed:
            if media_type == "svg":
                with path.open() as f:
                    svg_data = f.read()
                # Extract JUST the <svg> markup with no <XML> tag
                import xml.etree.ElementTree as ET

                tree = ET.fromstring(svg_data)
                just_svg_tag = ET.tostring(tree)
                return svg_group.render(
                    x=self.x, y=self.y, scale=self.scale, content=just_svg_tag
                )
            else:
                encoded_img = base64.b64encode(open(self.href, "rb").read())
                path = f"data:image/{media_type};base64,{encoded_img.decode('utf-8')}"

        return svg_image.render(
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            href=path,
        )


class Rect(Element):
    """SVG <rect> (rectangle) element."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def render(self):
        """Return SVG markup

        :return: SVG markup
        :rtype: string
        """
        return svg_rect.render(
            x=self.x,
            y=self.y,
            scale=self.scale,
            uid=uuid.uuid1(),
            **self.config,
        )


class Path(Element):
    """SVG <path> element"""

    def __init__(self, definition, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.definition = definition

    def render(self):
        return svg_path.render(d=self.definition, scale=self.scale, **self.config)


class Text(Element):
    """SVG <text> element."""

    def __init__(self, text_content, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text_content = text_content

    def render(self):
        """create an SVG <text> tag."""
        return svg_text.render(
            text_content=self.text_content,
            x=self.x,
            y=self.y,
            scale=self.scale,
            **self.config,
        )


class TextBlock(Element):
    """SVG <text> element."""

    def __init__(self, text_content, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if type(text_content) == str:
            # attempt to split on '\n' to make a list
            text_content = text_content.split("\n")
        self.text_content = text_content

    def render(self):
        translate = [self.x, self.y]
        if self.scale.x == -1:
            translate[0] = self.x + self.width
        if self.scale.y == -1:
            translate[1] = self.y + self.height
        """create an SVG <text> tag."""
        return svg_textblock.render(
            text_content=self.text_content,
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            scale=self.scale,
            translate=" ".join(str(coord) for coord in translate),
            **self.config,
        )


class Label(Element):
    """SVG <text> and <rect> markup in a single element."""

    def __init__(self, text_content, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text_content = text_content
        self.width = self.config["rect"]["width"]
        self.height = self.config["rect"]["height"]

    def render(self):
        """Render SVG component."""
        return svg_label.render(
            text_content=self.text_content,
            x=self.x,
            y=self.y,
            scale=self.scale,
            **self.config,
        )



import io
import math
import random
from pathlib import Path

from . import file_manager
from . import components


def luminace(color_component):
    """Luminance of an individual Red, Green, or Blue, color component.

    :param color_component: Value between 0 and 255 (inclusive)
    :type color_component: int
    :return: Luminance value of the color component
    :rtype: float
    """
    i = float(color_component) / 255

    if i < 0.03928:
        return i / 12.92
    else:
        return ((i + 0.055) / 1.055) ** 2.4


def relative_luminance(rgb):
    """Normalised luminance value of an RGB color.

    :param rgb: Tuple (or List) representing RGB value.
    :type rgb: tuple
    :return: Value between 0 and 1.
    :rtype: float
    """
    return (
        0.2126 * luminace(rgb[0])
        + 0.7152 * luminace(rgb[1])
        + 0.0722 * luminace(rgb[2])
    )


def random_contrasting_rgb(ref_color):
    """Generate a psudo-random color that, compared to 'ref_color', has a contrast ratio greater than 3. This contrast value is the minimum value to pass WCAG AA contrast recommendation for UI components.

    :param ref_color: Tuple (or List) representing RGB value.
    :type ref_color: tuple
    :return: Tuple representing an RGB color.
    :rtype: tuple
    """
    contrast = 0
    while contrast < 3:
        rgb = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
        lt = ref_color if sum(ref_color) > sum(rgb) else rgb
        dk = ref_color if sum(ref_color) < sum(rgb) else rgb
        contrast = (relative_luminance(lt) + 0.05) / (relative_luminance(dk) + 0.05)
    return rgb


def find_children_by_type(component, target_type):
    """Recursively find all children of the component and it's decendents by type.

    :param component: Component instance to start seach from
    :type component: class instance
    :param target_type: class  to match with instances
    :type target_type: class
    :return: All instances of type 'target_type' that are descendents of 'component'
    :rtype: list
    """
    results = []
    try:
        for c in component.children:
            if isinstance(c, target_type):
                results.append(c)
            results += find_children_by_type(c, target_type)
    except AttributeError:
        pass
        # No children

    return results


def default_css(diagram):
    """Generate a stylesheet using metrics from a diagram. Various styles are tailored by making a *best-guess* based on diagram component dimensions or a *lucky-guess* filtered by preset criteria. The output should be considered a boot-strapping step to styling a diagram ...unless you feel lucky!

    :param diagram: The Diagram object requiring styling
    :type diagram: Diagram
    :return: content of a css stylesheet with all required styles to display a diagram.
    :rtype: str
    """
    # Extract css class tags from PinLabels
    pinlabels = find_children_by_type(diagram, components.PinLabel)
    pinlabel_tags = list(
        set([tag for label in pinlabels for tag in label.tags.split(" ")])
    )
    # Remove config tag (common to all PinLabels)
    try:
        pinlabel_tags.remove(diagram.cfg.get("pinlabel", {}).get("tag", ""))
    except ValueError:
        pass

    label_font_size = diagram.cfg.get("pinlabel", {}).get(
        "font_size", math.floor(diagram.cfg["pinlabel"]["label"]["height"] * (3 / 5))
    )
    label_text_color = tuple(
        diagram.cfg.get("pinlabel", {}).get("text", {}).get("color", (255, 255, 255))
    )

    diagram.cfg["pinlabel"]["_categories"] = {
        label: random_contrasting_rgb(label_text_color) for label in pinlabel_tags
    }

    return stylesheet.render(diagram.cfg)



from pinout import diagram

# Custom pinlabel configuration
long_label_config = {"label": {"rect": {"width": 108}}}

# Pin label information grouped into dicts that render sets of labels
pin_headers = [
    {
        # LHS header
        "x": 16,
        "y": 100,
        "pitch": (0, 30),
        "offset": (-58, 0),
        "labels": [
            [("Vcc", "pwr", long_label_config)],
            [("1", "gpio"), ("A1", "analog")],
            [("2", "gpio"), ("PWM", "pwm", {"offset": (66, 0)})],
        ],
    },
    {
        # RHS header
        "x": 204,
        "y": 100,
        "pitch": (0, 30),
        "offset": (58, 0),
        "labels": [
            [("GND", "pwr", long_label_config)],
            [("7", "gpio"), ("A3", "analog")],
            [("6", "gpio"), ("A2", "analog"), ("PWM", "pwm")],
        ],
    },
    {
        # Lower header
        "x": 65,
        "y": 244,
        "pitch": (30, 0),
        "offset": (-107, 40),
        "labels": [
            [("RST", "pwr", long_label_config)],
            [("3", "gpio"), ("ADC0", "analog")],
            [("4", "gpio"), ("ADC1", "analog"), ("PWM", "pwm")],
            [("5", "gpio"), ("PWM", "pwm", {"offset": (66, 0)})],
        ],
    },
]

# Create a new diagram
diagram = diagram.Diagram()

# Override default config with user supplied values
diagram.add_config("quick_start_config.yaml")

# Add an image
diagram.add_image("quick_start_hardware.png", width=220, height=260, embed=True)

# Add a legend. Note, categories are documented in config.yaml
diagram.add_legend(x=250, y=200)

# Create pinout labels
for header in pin_headers:
    diagram.add_pinlabelset(**header)

# Export the finished diagram
diagram.export("quick_start_diagram.svg", overwrite=True)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name="pinout",
    version="0.0.9",
    author="John Newall",
    author_email="john@johnnewall.com",
    description="Generate graphical pinout references for electronic hardware.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/j0ono0/pinout",
    project_urls={
        "Bug Tracker": "https://github.com/j0ono0/pinout/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(include=["pinout", "pinout.*"]),
    python_requires=">=3.6",
    install_requires=["Jinja2", "pyyaml"],
    include_package_data=True,
)